// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of scatter
//
// scatter is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// scatter is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <iomanip>
#include <iostream>
#include <string>

#include <paramxx/ParFile.h>
#include <suplib/debug.h>
#include <tracefctxx/TraceFct.h>
#include <BinaryPipe/BinaryPipe.h>
#include <Exception/Exception.h>

#include "MirrorDatabase.h"
#include "MZ.h"

#include "Photon.h"
#include "Random.h"
#include "ScatterDatabase.h"

void process( ParFile& pf ) throw ( Exception ) {

  TraceFct tf( "process" );

  try {
    bool dump_offzone      = dbflag( "offzone" );
    bool dump_offzone2     = dbflag( "offzone2" );

    bool in_plane          = pf.pgetb( "in_plane" );
    bool out_of_plane      = pf.pgetb( "out_of_plane" );
    int  iter_limit        = pf.pgeti( "scat_iter_lim" );

    double theta_def       = ARCSEC2RAD(pf.pgetd( "theta" ));
    double phi_def         = ARCSEC2RAD(pf.pgetd( "phi" ));
    double max_cum_prob    = pf.pgetd( "max_cumulative_probability" );
    double min_cum_prob    = pf.pgetd( "min_cumulative_probability" );

    // Initialize the mirror position and scatter databases...
    MirrorDatabase mirrordb( pf.pgetstring( "mirror_db" ), 
			     pf.pgeti( "surf_no" ),
			     pf.pgeti( "shell" ) ) ;

    ScatterDatabase scatterdb( pf.pgetstring( "scatter_db" ), 
			       pf.pgeti( "surf_no" ),
			       pf.pgeti( "shell" ), 
			       mirrordb.z0( ) );


    struct DataPacket dp[3] = { { "theta", BPDType_double, NULL }, 
				{ "phi", BPDType_double, NULL }, 
				{ NULL, BPDType_double, NULL } };

    // Initialize I/O...
    BinaryPipe<Photon> binarypipe( pf.pgetstring( "input" ).c_str( ), pf.pgetstring( "output" ).c_str( ), NULL, dp );
    Photon *photon = static_cast<Photon*>(binarypipe);
    
    // don't output debug fields if we're not asked to
    if ( !dbflag( "theta" ) )
	photon->dpktf_delete( "theta" );

    if ( !dbflag( "phi" ) )
	photon->dpktf_delete( "phi" );

    unsigned int *surf_no;
    unsigned int surf_missed_mask = 0;
    if ( (surf_no = photon->get_srfno( )) ) {
      surf_missed_mask = 1 << (*surf_no - 1);
    }

    DVector3 *position        = photon->get_position( );
    DVector3 *direction       = photon->get_direction( );
    DVector3 *surface_norm    = photon->get_surface_norm( );
    double *energy            = photon->get_energy( );
    unsigned int *id          = photon->get_id( );
    unsigned int *surf_missed = photon->get_surf_missed( );
    double *theta            = photon->get_theta( );
    double *phi              = photon->get_phi( );

    // Initialize random number generator...
    Random randy( pf.pgeti( "seed1" ), pf.pgeti( "seed2" ), pf.pgeti( "block" ) );

    // Read photons...
    unsigned long n_offzone         = 0;
    unsigned long n_ip_extrapolated = 0;
    unsigned long n_op_extrapolated = 0;
    unsigned long n_ghosts          = 0;
    unsigned long n_surface_scat    = 0;
    unsigned long n_backscattered   = 0;

    while ( binarypipe.read_dpkts( ) ) {
      // If it hit the last surface let's try to scatter it...
      if ( ! ( surf_missed_mask & *surf_missed ) ) { 

	// Look at Z position of photon and select an appropriate
	// mirror zone...
	MZ *zone = scatterdb.findZone( position->z );
	
	// If this photon fell outside all mirror zones...
	if ( ! zone ) {
	  
	  if ( dump_offzone2 ) {
	    
	    double diff;
	    if ( position->z < scatterdb.zones( )[0]->zmin() ) 
	      diff = scatterdb.zones( )[0]->zmin() - position->z;
	    else
	      diff = position->z - scatterdb.zones( )[scatterdb.zones( ).size( )-1]->zmax();
	    
	    tf.message( "%u:  not in zone.  z=%f (%f,%f) diff = %f\n", *id, position->z,
			scatterdb.zones( )[0]->zmin(), 
			scatterdb.zones( )[scatterdb.zones( ).size( )-1]->zmax(), diff );
	    
	  }
	  
	  n_offzone++;
	  
	  if ( dump_offzone )
	    tf.message( "%u:  not in zone.  z = %f\n", *id, position->z );
	  
	  continue;
	  
	}
	
	// YOU HAVEN'T IMPLEMENTED THE skip ray if scattering is not
	// requested LOGIC!!!  is this option necessary?
	// skip ray if scattering is not requested 
	// if ( ! zone->flag || NULL == zone->scat_table )
	//   continue;
	
	// Finally, scatter the photon and, if it has scattered into
	// the surface or backscattered, absorb it.

	int status = zone->scatter( in_plane, out_of_plane, iter_limit,
				    theta_def, phi_def, min_cum_prob, max_cum_prob,
				    *id, *energy, *surface_norm, *direction, *theta, *phi );

	if ( status & MZ::SCATTERED_INTO_SURFACE )
	{
	  tf.message( "(%u): scattered into surface\n", *id );
	  n_surface_scat++;
	  continue;
	}

	if ( status & MZ::BACKSCATTERED )
	{
	  tf.message( "(%u): backscattered\n", *id );
	  n_backscattered++;
	  continue;
	}

	if ( status & MZ::IP_EXTRAPOLATED )
	{
	  n_ip_extrapolated++;

	  if ( dbflag( "extrapolated" ) ) 
	    tf.message( "(%u): in-plane extrapolated\n", *id );
	}

	if ( status & MZ::OP_EXTRAPOLATED ) 
	{
	  n_op_extrapolated++;

	  if ( dbflag( "extrapolated" ) ) 
	    tf.message( "(%u): out-of-plane extrapolated\n", *id );
	}

      // Else it missed the last surface and we should count it as a ghost...
      } else {

	n_ghosts++;

      }

      // Write photons...
      if ( binarypipe.write_dpkt( ) )
	throw( Exception( bpipe_strerror( bpipe_errno ) ) );

    }

    tf.message( "summary:  in-plane extrapolated:  %d\n", n_ip_extrapolated );
    tf.message( "summary:  out-of-plane extrapolated:  %d\n", n_op_extrapolated );
    tf.message( "summary:  off zone:  %d\n", n_offzone );
    tf.message( "summary:  scattered into surface:  %d\n", n_surface_scat );
    tf.message( "summary:  backscattered:  %d\n", n_backscattered );
    tf.message( "summary:  ghosts:  %d\n", n_ghosts );

  } catch ( Exception& E ) {

    throw( E );

  } catch ( exception& e ) {

    throw( Exception( e.what( ) ) );

  }
}
